"""
App constants
"""

# Standard Library
import re

# Django
from django.utils.text import slugify

# AA Intel Tool
from aa_intel_tool import __version__

# All internal URLs need to start with this prefix
INTERNAL_URL_PREFIX = "-"

SUPPORTED_INTEL_TYPES = [
    {
        "name": "Fleet Composition",
        "parser": "fleetcomp",
        "pattern": re.compile(
            r"(?im)^([a-zA-Z0-9 -_]{3,37})[\t](.*)[\t](.*)[\t](.*)[\t](.*)[\t]([0-5] - [0-5] - [0-5])([\t](.*))?$"  # pylint: disable=line-too-long
        ),
    },
    {
        "name": "D-Scan",
        "parser": "dscan",
        "pattern": re.compile(r"(?im)^(\d+)[\t](.*)[\t](.*)[\t](-|(.*) (km|m|AU)) ?$"),
    },
    {
        "name": "Chat List",
        "parser": "chatlist",
        "pattern": re.compile(r"(?im)^[a-zA-Z0-9\u0080-\uFFFF -_]{3,37}$"),
    },
]


VERBOSE_NAME = "AA Intel Tool"

verbose_name_slugified: str = slugify(VERBOSE_NAME, allow_unicode=True)
github_url: str = "https://github.com/ppfeufer/aa-intel-tool"

USER_AGENT = f"{verbose_name_slugified} v{__version__} {github_url}"



"""
Our URL config
"""

# Django
from django.urls import include, path

# AA Intel Tool
from aa_intel_tool.constants import INTERNAL_URL_PREFIX
from aa_intel_tool.views import ajax, general

app_name: str = "aa_intel_tool"


app_urls = [
    path(route="", view=general.index, name="intel_tool_index"),
    path(route="scan/<str:scan_hash>/", view=general.scan, name="intel_tool_scan"),
]

ajax_urls = [
    path(
        route="get-scan-data/<str:scan_hash>/<str:scan_section>/",
        view=ajax.get_scan_data,
        name="ajax_get_scan_data",
    ),
]

# Put it all together
urlpatterns = [
    # Ajax
    path(f"{INTERNAL_URL_PREFIX}/ajax/", include(ajax_urls)),
    # App URLs
    path("", include(app_urls)),
]



# Generated by Django 4.0.10 on 2023-08-09 07:37

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aa_intel_tool", "0003_alter_scandata_scan"),
    ]

    operations = [
        migrations.AlterField(
            model_name="scandata",
            name="section",
            field=models.CharField(
                choices=[
                    ("invalid", "Invalid scan data"),
                    ("pilotlist", "Pilot list"),
                    ("corporationlist", "Corporation list"),
                    ("alliancelist", "Alliance list"),
                    ("shiptypes", "Ship types"),
                    ("shiplist", "Ship list"),
                    ("shiplist_on_grid", "Ship list (on grid)"),
                    ("shiplist_off_grid", "Ship list (off grid)"),
                    ("structures_on_grid", "Structures (on grid)"),
                    ("deployables_on_grid", "Deployables (on grid)"),
                ],
                default="invalid",
                max_length=19,
                verbose_name="The scan section",
            ),
        ),
    ]



"""
Helper for our tests
"""

# Standard Library
import re
from pathlib import Path
from typing import List

# Third Party
from faker import Faker

# Django
from django.contrib.auth.models import User
from django.template import Context, Template

# Alliance Auth
from allianceauth.tests.auth_utils import AuthUtils

fake = Faker()


def create_fake_user(  # pylint: disable=too-many-arguments
    character_id: int,
    character_name: str,
    corporation_id: int = None,
    corporation_name: str = None,
    corporation_ticker: str = None,
    permissions: List[str] = None,
    **kwargs,
) -> User:
    """
    Create a fake user including its main character and (optional) permissions.
    :param character_id:
    :param character_name:
    :param corporation_id:
    :param corporation_name:
    :param corporation_ticker:
    :param permissions:
    :param kwargs:
    :return:
    """

    username = re.sub(pattern=r"[^\w\d@.+-]", repl="_", string=character_name)
    user = AuthUtils.create_user(username=username)

    if not corporation_id:
        corporation_id = 2001
        corporation_name = "Wayne Technologies Inc."
        corporation_ticker = "WTE"

    alliance_id = kwargs.get("alliance_id", 3001)
    alliance_name = (
        kwargs.get("alliance_name", "Wayne Enterprises")
        if alliance_id is not None
        else ""
    )

    AuthUtils.add_main_character_2(
        user=user,
        name=character_name,
        character_id=character_id,
        corp_id=corporation_id,
        corp_name=corporation_name,
        corp_ticker=corporation_ticker,
        alliance_id=alliance_id,
        alliance_name=alliance_name,
    )

    if permissions:
        perm_objs = [
            AuthUtils.get_permission_by_name(perm=perm) for perm in permissions
        ]
        user = AuthUtils.add_permissions_to_user(perms=perm_objs, user=user)

    return user


def get_or_create_fake_user(*args, **kwargs) -> User:
    """
    Same as create_fake_user but will not fail when user already exists.
    """

    if len(args) > 1:
        character_name = args[1]
    elif "character_name" in kwargs:
        character_name = kwargs["character_name"]
    else:
        ValueError("character_name is not defined")

    username = character_name.replace("'", "").replace(" ", "_")

    try:
        return User.objects.get(username=username)
    except User.DoesNotExist:  # pylint: disable=no-member
        return create_fake_user(*args, **kwargs)


def render_template(string, context=None):
    """
    Helper to render templates
    :param string:
    :param context:
    :return:
    """

    context = context or {}
    context = Context(dict_=context)

    return Template(template_string=string).render(context=context)


def load_chatscan_txt() -> str:
    """
    Loading chatscan.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/chatscan.txt").read_text()


def load_chatscan_faulty_txt() -> str:
    """
    Loading chatscan-faulty.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/chatscan-faulty.txt").read_text()


def load_dscan_txt() -> str:
    """
    Loading dscan.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/dscan.txt").read_text()


def load_fleetcomp_txt() -> str:
    """
    Loading fleetcomp.txt

    :return:
    :rtype:
    """

    return (Path(__file__).parent / "test-data/fleetcomp.txt").read_text()



# Generated by Django 4.0.10 on 2023-08-04 15:02

# Django
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Scan",
            fields=[
                (
                    "hash",
                    models.CharField(
                        editable=False,
                        max_length=30,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name="The hash of the scan.",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        auto_now_add=True,
                        db_index=True,
                        verbose_name="Date and time the scan was created.",
                    ),
                ),
                ("raw_data", models.TextField(verbose_name="The original scan data.")),
                (
                    "processed_data",
                    models.JSONField(
                        default=dict, verbose_name="The processed scan data."
                    ),
                ),
                (
                    "scan_type",
                    models.CharField(
                        choices=[
                            ("invalid", "Invalid scan data"),
                            ("dscan", "D-Scan"),
                            ("fleetcomp", "Fleet Composition"),
                            ("chatlist", "Chat List"),
                        ],
                        default="invalid",
                        max_length=9,
                        verbose_name="The scan type.",
                    ),
                ),
            ],
            options={
                "verbose_name": "scan",
                "verbose_name_plural": "scans",
                "default_permissions": (),
            },
        ),
    ]



"""
Django admin integration
"""

# Django
from django.contrib import admin

# AA Intel Tool
from aa_intel_tool.models import Scan


class BaseReadOnlyAdminMixin:
    """
    Base "Read Only" mixin for admin models
    """

    actions = None  # Removes the default delete action.

    def has_add_permission(self, request):  # pylint: disable=unused-argument
        """
        Has "add" permissions

        :param request:
        :type request:
        :return:
        :rtype:
        """

        return False

    def has_change_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "change" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False

    def has_delete_permission(
        self, request, obj=None  # pylint: disable=unused-argument
    ):
        """
        Has "delete" permissions

        :param request:
        :type request:
        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        return False


@admin.register(Scan)
class ScanAdmin(BaseReadOnlyAdminMixin, admin.ModelAdmin):
    """
    Scan Admin
    """

    list_display = ("hash", "created")



"""
App settings
"""

# Alliance Auth (External Libs)
from app_utils.app_settings import clean_setting


class AppSettings:  # pylint: disable=too-few-public-methods
    """
    App settings from local.py
    """

    # Enable or disable the chat scan module
    # Disabled by default
    INTELTOOL_ENABLE_MODULE_CHATSCAN = clean_setting(
        name="INTELTOOL_ENABLE_MODULE_CHATSCAN", default_value=False, required_type=bool
    )

    # Enable or disable the d-scan module
    # Enabled by default
    INTELTOOL_ENABLE_MODULE_DSCAN = clean_setting(
        name="INTELTOOL_ENABLE_MODULE_DSCAN", default_value=True, required_type=bool
    )

    # Enable or disable the fleet composition module
    # Enabled by default
    INTELTOOL_ENABLE_MODULE_FLEETCOMP = clean_setting(
        name="INTELTOOL_ENABLE_MODULE_FLEETCOMP",
        default_value=False,
        required_type=bool,
    )

    # Scan retention time
    # Sets the time in days for how long the scans will be kept in the database.
    # Set to 0 to keep scans indefinitely.
    INTELTOOL_SCAN_RETENTION_TIME = clean_setting(
        name="INTELTOOL_SCAN_RETENTION_TIME", default_value=30, required_type=int
    )

    # Set the maximum number of pilots allowed per chat scan
    # Set to 0 for no limit.
    INTELTOOL_CHATSCAN_MAX_PILOTS = clean_setting(
        name="INTELTOOL_CHATSCAN_MAX_PILOTS", default_value=500, required_type=int
    )



"""
Chat list parser
"""

# Standard Library
from typing import Union

# Django
from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext

# Alliance Auth
from allianceauth.eveonline.evelinks import dotlan, eveimageserver, evewho, zkillboard
from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveEntity

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AppSettings
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.helper.eve_character import get_or_create_character
from aa_intel_tool.models import Scan, ScanData
from aa_intel_tool.parser.helper.db import safe_scan_to_db

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def _parse_alliance_info(eve_character: EveCharacter) -> dict:
    """
    Parse the alliance information from an EveCharacter

    :param eve_character:
    :type eve_character:
    :return:
    :rtype:
    """

    if eve_character.alliance_id is None:
        alliance_info = {
            "id": 1,
            "name": "",
            "ticker": "",
            "logo": eveimageserver.alliance_logo_url(alliance_id=1, size=32),
        }
    else:
        alliance_info = {
            "id": eve_character.alliance_id,
            "name": eve_character.alliance_name,
            "ticker": eve_character.alliance_ticker,
            "logo": eveimageserver.alliance_logo_url(
                alliance_id=eve_character.alliance_id, size=32
            ),
            "dotlan": dotlan.alliance_url(eve_character.alliance_name),
            "zkillboard": zkillboard.alliance_url(eve_character.alliance_id),
        }

    return alliance_info


def _parse_corporation_info(eve_character: EveCharacter) -> dict:
    """
    Parse the corporation information from an EveCharacter

    :param eve_character:
    :type eve_character:
    :return:
    :rtype:
    """

    return {
        "id": eve_character.corporation_id,
        "name": eve_character.corporation_name,
        "ticker": eve_character.corporation_ticker,
        "logo": eveimageserver.corporation_logo_url(
            corporation_id=eve_character.corporation_id, size=32
        ),
        "dotlan": dotlan.corporation_url(eve_character.corporation_name),
        "zkillboard": zkillboard.corporation_url(eve_character.corporation_id),
    }


def _parse_character_info(eve_character: EveCharacter) -> dict:
    """
    Parse the character information from an EveCharacter

    :param eve_character:
    :type eve_character:
    :return:
    :rtype:
    """

    return {
        "id": eve_character.character_id,
        "name": eve_character.character_name,
        "portrait": eveimageserver.character_portrait_url(
            character_id=eve_character.character_id, size=32
        ),
        "evewho": evewho.character_url(eve_character.character_id),
        "zkillboard": zkillboard.character_url(eve_character.character_id),
    }


def _parse_chatscan_data(eve_characters: QuerySet[EveCharacter]) -> dict:
    """
    Parse the chat scan data and return character information,
    corporation information and alliance information for each character

    :param eve_characters:
    :type eve_characters:
    :return:
    :rtype:
    """

    counter = {}
    alliance_info = {}
    corporation_info = {}
    character_info = {}

    for eve_character in eve_characters:
        eve_character__alliance_name = "Unaffiliated"

        if eve_character.alliance_name is not None:
            eve_character__alliance_name = eve_character.alliance_name

        if eve_character__alliance_name not in counter:
            counter[eve_character__alliance_name] = 0

        if eve_character.corporation_name not in counter:
            counter[eve_character.corporation_name] = 0

        # Alliance Info
        if eve_character__alliance_name not in alliance_info:
            alliance_info[eve_character__alliance_name] = _parse_alliance_info(
                eve_character=eve_character
            )

        # Corporation Info
        if eve_character.corporation_name not in corporation_info:
            corporation_info[eve_character.corporation_name] = _parse_corporation_info(
                eve_character=eve_character
            )
            corporation_info[eve_character.corporation_name][
                "alliance"
            ] = alliance_info[eve_character__alliance_name]

        # Character Info
        character_info[eve_character.character_name] = _parse_character_info(
            eve_character=eve_character
        )
        character_info[eve_character.character_name]["corporation"] = corporation_info[
            eve_character.corporation_name
        ]
        character_info[eve_character.character_name]["alliance"] = alliance_info[
            eve_character__alliance_name
        ]

        # Update the counter
        counter[eve_character__alliance_name] += 1
        alliance_info[eve_character__alliance_name]["count"] = counter[
            eve_character__alliance_name
        ]

        counter[eve_character.corporation_name] += 1
        corporation_info[eve_character.corporation_name]["count"] = counter[
            eve_character.corporation_name
        ]

    # Sort and clean up the dicts
    cleaned_pilot_data = [
        character
        for (
            character_name,  # pylint: disable=unused-variable
            character,
        ) in sorted(character_info.items())
    ]
    cleaned_corporation_data = [
        corporation
        for (
            corporation_name,  # pylint: disable=unused-variable
            corporation,
        ) in sorted(corporation_info.items())
    ]
    cleaned_alliance_data = [
        alliance
        for (
            alliance_name,  # pylint: disable=unused-variable
            alliance,
        ) in sorted(alliance_info.items())
    ]

    return {
        "pilots": cleaned_pilot_data,
        "corporations": cleaned_corporation_data,
        "alliances": cleaned_alliance_data,
    }


def parse(scan_data: list, safe_to_db: bool = True) -> Union[Scan, dict]:
    """
    Parse chat list

    :param scan_data:
    :type scan_data:
    :param safe_to_db:
    :type safe_to_db:
    :return:
    :rtype:
    """

    message = _("The chat list module is currently disabled.")

    if AppSettings.INTELTOOL_ENABLE_MODULE_CHATSCAN is True:
        logger.debug(msg=f"{len(scan_data)} name(s) to work through …")

        pilots_in_scan = len(scan_data)
        max_allowed_pilots = AppSettings.INTELTOOL_CHATSCAN_MAX_PILOTS

        if 0 < max_allowed_pilots < pilots_in_scan:
            logger.debug(
                msg=(
                    f"Number of pilots in scan ({pilots_in_scan}) exceeds the maximum "
                    f"allowed number ({max_allowed_pilots}). Throwing a tantrum …"
                )
            )

            raise ParserError(
                message=ngettext(
                    singular=f"Chat scans are currently limited to a maximum of {max_allowed_pilots} pilot per scan. Your list of pilots exceeds this limit.",  # pylint: disable=line-too-long
                    plural=f"Chat scans are currently limited to a maximum of {max_allowed_pilots} pilots per scan. Your list of pilots exceeds this limit.",  # pylint: disable=line-too-long
                    number=max_allowed_pilots,
                )
            )

        # Check if we have to bother Eve Universe or if we have all characters already
        fetch_from_eveuniverse = False
        try:
            eve_characters = EveCharacter.objects.filter(character_name__in=scan_data)
        except EveCharacter.DoesNotExist:  # pylint: disable=no-member
            fetch_from_eveuniverse = True
        else:
            if len(scan_data) != eve_characters.count():
                fetch_from_eveuniverse = True

        if fetch_from_eveuniverse:
            try:
                eve_character_ids = (
                    EveEntity.objects.fetch_by_names_esi(names=scan_data)
                    .filter(category=EveEntity.CATEGORY_CHARACTER)
                    .values_list("id", flat=True)
                )
            except EveEntity.DoesNotExist as exc:  # pylint: disable=no-member
                message = _(
                    "Something went wrong while fetching the character information from ESI."
                )

                raise ParserError(message=message) from exc

            logger.debug(
                msg=f"Got {len(eve_character_ids)} ID(s) back from Eve Universe …"
            )

            # In case the name does not belong to an Eve character,
            # EveEntity returns an empty object
            if len(eve_character_ids) == 0:
                message = _("Character unknown to ESI.")

                raise ParserError(message=message)

            eve_characters = get_or_create_character(character_ids=eve_character_ids)

        logger.debug(
            msg=f"Got {len(eve_characters)} EveCharacter object(s) back from AA …"
        )

        # Parse the data
        parsed_chatscan = _parse_chatscan_data(eve_characters=eve_characters)

        parsed_data = {
            "pilots": {
                "section": ScanData.Section.PILOTLIST,
                "data": parsed_chatscan["pilots"],
            },
            "corporations": {
                "section": ScanData.Section.CORPORATIONLIST,
                "data": parsed_chatscan["corporations"],
            },
            "alliances": {
                "section": ScanData.Section.ALLIANCELIST,
                "data": parsed_chatscan["alliances"],
            },
        }

        if safe_to_db is False:
            return parsed_data

        return safe_scan_to_db(scan_type=Scan.Type.CHATLIST, parsed_data=parsed_data)

    raise ParserError(message=message)



"""
Fleet composition parser
"""

# pylint: disable=unreachable

# Standard Library
import re

# Django
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AppSettings
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.models import Scan
from aa_intel_tool.parser.helper.db import safe_scan_to_db
from aa_intel_tool.parser.module.chatlist import parse as parse_pilots

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def parse(scan_data: list) -> Scan:
    """
    Parse chat list

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    message = _("The fleet composition module is currently disabled.")

    if AppSettings.INTELTOOL_ENABLE_MODULE_FLEETCOMP is True:
        raise ParserError(
            message=_("The fleet composition module is not yet finished, be patient …")
        )

        logger.debug(msg=scan_data)

        # Let's split this list up
        #
        # [0] => Pilot Name
        # [1] => System
        # [2] => Ship Class
        # [3] => Ship Type
        # [4] => Position in Fleet
        # [5] => Skills (FC - WC - SC)
        # [6] => Wing Name / Squad Name
        pilots = {"list": [], "flying": []}
        lines = []
        for entry in scan_data:
            line = re.split(pattern=r"\t+", string=entry.rstrip("\t"))

            if len(line) == 6:
                line.append("")

            pilots["list"].append(line[0])
            pilots["flying"].append([line[0], line[2], line[1]])
            lines.append(line)

        for line in lines:
            logger.debug(line[0] + " » " + line[6])

        logger.debug(msg=pilots)

        pilotlist = parse_pilots(scan_data=pilots["list"], safe_to_db=False)

        logger.debug(msg=pilotlist)

        parsed_data = {
            "general_information": None,
            "ship_classes": None,
            "ship_types": None,
            "composition": None,
            "pilots": pilotlist["pilots"] if pilotlist is not None else None,
            "corporations": pilotlist["corporations"],
            "alliances": pilotlist["alliances"],
        }

        return safe_scan_to_db(scan_type=Scan.Type.FLEETCOMP, parsed_data=parsed_data)

    raise ParserError(message=message)



"""
Tests for the template tags
"""


# Django
from django.test import TestCase

# AA Intel Tool
from aa_intel_tool import __version__
from aa_intel_tool.tests.utils import render_template


class TestForumVersionedStatic(TestCase):
    """
    Tests for inteltool_versioned_static template tag
    """

    def test_versioned_static(self):
        """
        Test should return static URL string with a version
        :return:
        """

        context = {"version": __version__}

        rendered_template = render_template(
            string=(
                "{% load inteltool_versioned_static %}"
                "{% aa_intel_tool_static 'aa_intel_tool/css/aa-intel-tool.min.css' %}"
            ),
            context=context,
        )

        self.assertEqual(
            rendered_template,
            f'/static/aa_intel_tool/css/aa-intel-tool.min.css?v={context["version"]}',
        )



# Generated by Django 4.0.10 on 2023-08-06 09:03

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aa_intel_tool", "0001_initial"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="scan",
            name="processed_data",
        ),
        migrations.CreateModel(
            name="ScanData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "section",
                    models.CharField(
                        choices=[
                            ("invalid", "Invalid scan data"),
                            ("pilotlist", "Pilot list"),
                            ("corporationlist", "Corporation list"),
                            ("alliancelist", "Alliance list"),
                        ],
                        default="invalid",
                        max_length=15,
                        verbose_name="The scan section",
                    ),
                ),
                (
                    "processed_data",
                    models.JSONField(
                        default=dict, verbose_name="The processed scan section data"
                    ),
                ),
                (
                    "scan",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="scan_data",
                        to="aa_intel_tool.scan",
                        verbose_name="Scan",
                    ),
                ),
            ],
            options={
                "verbose_name": "scan data",
                "verbose_name_plural": "scan data",
                "default_permissions": (),
            },
        ),
    ]



"""
D-Scan parser
"""

# Standard Library
import re

# Django
from django.db.models import QuerySet
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks import eveimageserver
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.constants import EveCategoryId
from eveuniverse.models import EveType

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AppSettings
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.models import Scan, ScanData
from aa_intel_tool.parser.helper.db import safe_scan_to_db

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def _parse_ship_information(eve_types: QuerySet, counter: dict) -> dict:
    ships = {"all": {}, "ongrid": {}, "offgrid": {}, "types": {}}

    eve_types_ships = eve_types.filter(
        eve_group__eve_category_id__exact=EveCategoryId.SHIP
    )

    for ship_id, ship_name, ship_type_id, ship_type_name in eve_types_ships:
        if ship_id in counter["all"]:
            ships["all"][ship_name] = {
                "id": ship_id,
                "name": ship_name,
                "type_id": ship_type_id,
                "type_name": ship_type_name,
                "type_name_sanitised": slugify(ship_type_name),
                "count": counter["all"][ship_id],
                "image": eveimageserver.type_icon_url(type_id=ship_id, size=32),
            }

            if ship_id in counter["ongrid"]:
                ships["ongrid"][ship_name] = {
                    "id": ship_id,
                    "name": ship_name,
                    "type_id": ship_type_id,
                    "type_name": ship_type_name,
                    "type_name_sanitised": slugify(ship_type_name),
                    "count": counter["ongrid"][ship_id],
                    "image": eveimageserver.type_icon_url(type_id=ship_id, size=32),
                }

            if ship_id in counter["offgrid"]:
                ships["offgrid"][ship_name] = {
                    "id": ship_id,
                    "name": ship_name,
                    "type_id": ship_type_id,
                    "type_name": ship_type_name,
                    "type_name_sanitised": slugify(ship_type_name),
                    "count": counter["offgrid"][ship_id],
                    "image": eveimageserver.type_icon_url(type_id=ship_id, size=32),
                }

    for ship_name, ship_info in ships["all"].items():
        if ship_info["type_name"] not in counter["type"]:
            counter["type"][ship_info["type_name"]] = 0

        counter["type"][ship_info["type_name"]] += ship_info["count"]

        if ship_info["type_name"] not in ships["types"]:
            ships["types"][ship_info["type_name"]] = {
                "name": ship_info["type_name"],
                "name_sanitised": slugify(ship_info["type_name"]),
            }

        ships["types"][ship_info["type_name"]]["count"] = counter["type"][
            ship_info["type_name"]
        ]

    # Sort and clean up the dicts
    cleaned_ships_all = [
        ship
        for (
            ship_name,  # pylint: disable=unused-variable
            ship,
        ) in sorted(ships["all"].items())
    ]
    cleaned_ships_ongrid = [
        ship
        for (
            ship_name,  # pylint: disable=unused-variable
            ship,
        ) in sorted(ships["ongrid"].items())
    ]
    cleaned_ships_offgrid = [
        ship
        for (
            ship_name,  # pylint: disable=unused-variable
            ship,
        ) in sorted(ships["offgrid"].items())
    ]
    cleaned_ships_types = [
        ship
        for (
            ship_name,  # pylint: disable=unused-variable
            ship,
        ) in sorted(ships["types"].items())
    ]

    return {
        "all": cleaned_ships_all,
        "ongrid": cleaned_ships_ongrid,
        "offgrid": cleaned_ships_offgrid,
        "types": cleaned_ships_types,
    }


def parse(scan_data: list) -> Scan:
    """
    Parse D-Scan

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    message = _("The D-Scan module is currently disabled.")

    if AppSettings.INTELTOOL_ENABLE_MODULE_DSCAN is True:
        counter = {"all": {}, "ongrid": {}, "offgrid": {}, "type": {}}
        eve_ids = {"all": [], "ongrid": [], "offgrid": []}

        # Let's split this list up
        #
        # [0] => Item ID
        # [1] => Name
        # [2] => Ship Class
        # [3] => Distance
        for entry in scan_data:
            line = re.split(pattern=r"\t+", string=entry.rstrip("\t"))
            entry_id = int(line[0])

            if entry_id not in counter["all"]:
                counter["all"][entry_id] = 0

            if line[3] == "-":
                if entry_id not in counter["offgrid"]:
                    counter["offgrid"][entry_id] = 0

                counter["offgrid"][entry_id] += 1
                eve_ids["offgrid"].append(entry_id)
            else:
                if entry_id not in counter["ongrid"]:
                    counter["ongrid"][entry_id] = 0

                counter["ongrid"][entry_id] += 1
                eve_ids["ongrid"].append(entry_id)

            counter["all"][entry_id] += 1

            eve_ids["all"].append(entry_id)

        eve_types = EveType.objects.bulk_get_or_create_esi(
            ids=set(eve_ids["all"]), include_children=True
        ).values_list("id", "name", "eve_group__id", "eve_group__name")

        # Parse the data
        ships = _parse_ship_information(eve_types=eve_types, counter=counter)

        parsed_data = {
            "shiptypes": {
                "section": ScanData.Section.SHIPTYPES,
                "data": ships["types"],
            },
            "all": {
                "section": ScanData.Section.SHIPLIST,
                "data": ships["all"],
            },
            "ongrid": {
                "section": ScanData.Section.SHIPLIST_ON_GRID,
                "data": ships["ongrid"],
            },
            "offgrid": {
                "section": ScanData.Section.SHIPLIST_OFF_GRID,
                "data": ships["offgrid"],
            },
        }

        return safe_scan_to_db(scan_type=Scan.Type.DSCAN, parsed_data=parsed_data)

    raise ParserError(message=message)



"""
DB helper for our parser
"""

# AA Intel Tool
from aa_intel_tool.models import Scan, ScanData


def safe_scan_to_db(scan_type: Scan.Type, parsed_data: dict) -> Scan:
    """
    Saving scan data to the DB

    :param scan_type:
    :type scan_type:
    :param parsed_data:
    :type parsed_data:
    :return:
    :rtype:
    """

    # Creating a new Scan object
    new_scan = Scan(
        scan_type=scan_type,
    )
    new_scan.save()

    # Creating the associated ScanData objects
    for _, scan in parsed_data.items():
        ScanData(
            scan=new_scan,
            section=scan["section"],
            processed_data=scan["data"],
        ).save()

    # Return the Scan object
    return new_scan



# Generated by Django 4.0.10 on 2023-08-06 09:21

# Django
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("aa_intel_tool", "0002_remove_scan_processed_data_scandata"),
    ]

    operations = [
        migrations.AlterField(
            model_name="scandata",
            name="scan",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="scan_data",
                to="aa_intel_tool.scan",
                verbose_name="Scan",
            ),
        ),
    ]



"""
The tasks
"""

# Standard Library
from datetime import timedelta

# Third Party
from celery import shared_task

# Django
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.app_settings import AppSettings
from aa_intel_tool.models import Scan

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@shared_task
def housekeeping() -> None:
    """
    Remove scans older than AppSettings.INTELTOOL_SCAN_RETENTION_TIME day(s)

    :return:
    :rtype:
    """

    if AppSettings.INTELTOOL_SCAN_RETENTION_TIME > 0:
        logger.info(
            msg=f"Removing scans older than {AppSettings.INTELTOOL_SCAN_RETENTION_TIME} day(s)"  # pylint: disable=line-too-long
        )

        Scan.objects.filter(  # pylint: disable=no-member
            created__lte=timezone.now()
            - timedelta(days=AppSettings.INTELTOOL_SCAN_RETENTION_TIME)
        ).delete()



"""
Tests for our models
"""

# Django
from django.test import TestCase

# AA Intel Tool
from aa_intel_tool.models import Scan


class TestScan(TestCase):
    """
    Tests for the board model
    """

    def test_model_string_names(self):
        """
        Test model string name

        :return:
        :rtype:
        """

        scan = Scan(
            raw_data="Foobar",
        )
        scan.save()
        expected_hash = scan.hash
        self.assertEqual(first=str(scan), second=expected_hash)

        scan.raw_data = "BarFoo"
        scan.save()
        self.assertEqual(first=str(scan), second=expected_hash)

        scan.hash = ""
        scan.save()
        self.assertNotEqual(first=str(scan), second="")
        self.assertNotEqual(first=str(scan), second=expected_hash)

        scan.save()



"""
Testing the parsers
"""

# Django
from django.test import TestCase

# AA Intel Tool
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.parser.general import check_intel_type, parse_intel
from aa_intel_tool.tests.utils import (
    load_chatscan_faulty_txt,
    load_chatscan_txt,
    load_dscan_txt,
    load_fleetcomp_txt,
)


class TestParserGeneral(TestCase):
    """
    Testing parser.general.py
    """

    def test_check_intel_type_dscan(self):
        """
        Test should return 'dscan' as the expected intel type

        :return:
        :rtype:
        """

        form_data = load_dscan_txt()
        scan_data = str(form_data).splitlines()

        intel_type = check_intel_type(scan_data=scan_data)
        expected_intel_type = "dscan"

        self.assertEqual(first=intel_type, second=expected_intel_type)

    def test_check_intel_type_chatlist(self):
        """
        Test should return 'chatlist' as the expected intel type

        :return:
        :rtype:
        """

        form_data = load_chatscan_txt()
        scan_data = str(form_data).splitlines()

        intel_type = check_intel_type(scan_data=scan_data)
        expected_intel_type = "chatlist"

        self.assertEqual(first=intel_type, second=expected_intel_type)

    def test_check_intel_type_fleetcomp(self):
        """
        Test should return 'fleetcomp' as the expected intel type

        :return:
        :rtype:
        """

        form_data = load_fleetcomp_txt()
        scan_data = str(form_data).splitlines()

        intel_type = check_intel_type(scan_data=scan_data)
        expected_intel_type = "fleetcomp"

        self.assertEqual(first=intel_type, second=expected_intel_type)

    def test_check_intel_type_none(self):
        """
        Test should throw a ParserError as the expected intel type
        This happens when invalid data has been posted

        :return:
        :rtype:
        """

        form_data = load_chatscan_faulty_txt()
        scan_data = str(form_data).splitlines()

        expected_exception = ParserError
        expected_message = "A parser error occurred » No suitable parser found …"

        with self.assertRaises(ParserError):
            check_intel_type(scan_data=scan_data)

        with self.assertRaisesMessage(
            expected_exception=expected_exception, expected_message=expected_message
        ):
            check_intel_type(scan_data=scan_data)

    def test_parse_intel_with_invalid_form_data(self):
        """
        Test should return a ParserError as parsed intel data for invalid form data

        :return:
        :rtype:
        """

        form_data = load_chatscan_faulty_txt()

        expected_exception = ParserError
        expected_message = "A parser error occurred » No suitable parser found …"

        with self.assertRaises(ParserError):
            parse_intel(form_data=form_data)

        with self.assertRaisesMessage(
            expected_exception=expected_exception, expected_message=expected_message
        ):
            parse_intel(form_data=form_data)

    def test_parse_intel_empty_form_data(self):
        """
        Test should throw a ParserError as parsed intel data for empty form data

        :return:
        :rtype:
        """

        form_data = ""

        expected_exception = ParserError
        expected_message = "A parser error occurred » No data to parse …"

        with self.assertRaises(expected_exception=expected_exception):
            parse_intel(form_data=form_data)

        with self.assertRaisesMessage(
            expected_exception=expected_exception, expected_message=expected_message
        ):
            parse_intel(form_data=form_data)



"""
General parser functions
"""

# Standard Library
import re

# Django
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Intel Tool
from aa_intel_tool import __title__
from aa_intel_tool.constants import SUPPORTED_INTEL_TYPES
from aa_intel_tool.exceptions import ParserError
from aa_intel_tool.parser.module import chatlist, dscan, fleetcomp

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


def check_intel_type(scan_data: list) -> str:
    """
    Check which intel type we have

    :param scan_data:
    :type scan_data:
    :return:
    :rtype:
    """

    for intel_type in SUPPORTED_INTEL_TYPES:
        if all(
            re.match(pattern=intel_type["pattern"], string=string)
            for string in scan_data
        ):
            return intel_type["parser"]

    raise ParserError(message=_("No suitable parser found …"))


def parse_intel(form_data: str) -> str:
    """
    Parse intel

    :param form_data:
    :type form_data:
    :return:
    :rtype:
    """

    scan_data = form_data.splitlines()

    if len(scan_data) > 0:
        try:
            intel_type = check_intel_type(scan_data=scan_data)
        except ParserError as exc:
            raise ParserError(message=exc.message) from exc

        available_parser = {
            "dscan": dscan.parse,
            "chatlist": chatlist.parse,
            "fleetcomp": fleetcomp.parse,
        }

        if intel_type in available_parser:
            try:
                new_scan = available_parser[intel_type](scan_data=scan_data)
            except ParserError as exc:
                # Re-raise the Exception
                raise ParserError(message=exc.message) from exc

            # if new_scan is not None:
            new_scan.raw_data = form_data
            new_scan.save()

            return new_scan.hash

        raise ParserError(message=_("No suitable parser found …"))

    raise ParserError(message=_("No data to parse …"))
